-----------------------------------
-- FeatureTweezer module: FeatureIO
-- 2010-10 Wolfgang Scholz
-----------------------------------
-- This module handles several input formats for interoperability with other tools
module FeatureIO (printProductLineAbstractness, printReferentialAbstractness) where 

import GuiDSL
import IO
import System.IO.Unsafe
import FeatureStructure
import ProductLine
import qualified Data.Set
import Data.Set (Set)
import qualified RSF
import qualified Introduces
import qualified CompOrder
import qualified Data.Map
import Data.Map (Map)
import Maybe
import Control.Monad

showGuiDSLmodel :: String -> IO ()
showGuiDSLmodel path = do
    s <- readFile path
    let str = (show . parseGuiDSL . lexer . discardFormatting . removeComments) s
    putStr str
    putStr "\n"

readGuiDSLmodel :: String -> ([DProduction],[DConstr])
readGuiDSLmodel path = 
    (parseGuiDSL . lexer . discardFormatting . removeComments . unsafePerformIO)
    (readFile path)

-- transforms a GuiDSL model into the internal product line data type
guiDSLtoProductLine :: ([DProduction],[DConstr]) -> ProductLine
guiDSLtoProductLine (prs,cns) = (feats, extractFMfromGuiDSL feats prs cns)
    where feats = extractFeaturesFromGuiDSL prs

-- extracts an internal feature model out of a guiDSL file
extractFMfromGuiDSL :: Set Feature -> [DProduction] -> [DConstr] -> [DomainConstraint]
extractFMfromGuiDSL fs ps cs = --TODO empty product currently allowed
    (DConstraint [] [Must (getRootFeature fs ps)]) : 
    ((extractFMfromGuiDSLprod fs ps) ++ (extractFMfromGuiDSLconstr fs cs))
    where getRootFeature :: Set Feature -> [DProduction] -> Feature
          getRootFeature fs [] = error "guidsl model is empty"
          getRootFeature fs ((DProd s ps):_) = getFeatureByName fs s

-- extracts feature model domain rules out of a GuiDSL model
extractFMfromGuiDSLprod :: Set Feature -> [DProduction] -> [DomainConstraint]
extractFMfromGuiDSLprod fs [] = []
extractFMfromGuiDSLprod fs (p:prs) =  (extractFMprod fs p) ++ (extractFMfromGuiDSLprod fs prs)
    where
        extractFMprod :: Set Feature -> DProduction -> [DomainConstraint]
        extractFMprod fs (DProd s ps) = 
            (DConstraint [getFeatureByName fs s]
                (map (\p -> Must (getFeatureByName fs p)) (allThisLevelPats ps)))
            : ( [DConstraint [x] [MustNot y] | 
                    x <- (map (getFeatureByName fs) (allThisLevelPats ps)),
                    y <- (map (getFeatureByName fs) (allThisLevelPats ps)),
                    x /= y ]
              ++ (map (\p -> DConstraint [getFeatureByName fs p] 
                 [(Must (getFeatureByName fs s))]) (allThisLevelPats ps)) 
              ++ (extractFMpats fs ps) )
        --allThisLevelPats: list of all feature strings of the next level [used in refs]
        allThisLevelPats :: DPatterns -> [String]
        allThisLevelPats (DPatn (DSeriesPat s _)) = [s]
        allThisLevelPats (DPatn (DSinglePat (DOptTerm s))) = [s]
        allThisLevelPats (DPatns (DSeriesPat s _) ps) = s : (allThisLevelPats ps)
        allThisLevelPats (DPatns (DSinglePat (DOptTerm s)) ps) = s : (allThisLevelPats ps)
        --patterns - go deeper in the hieriarchy, searching for model constraints
        extractFMpats :: Set Feature -> DPatterns -> [DomainConstraint]
        extractFMpats fs (DPatn p) = (extractFMpat fs p)
        extractFMpats fs (DPatns p ps) = (extractFMpat fs p) ++ (extractFMpats fs ps)
        --single pattern
        extractFMpat :: Set Feature -> DPattern -> [DomainConstraint]
        extractFMpat fs (DSinglePat _) = [] --no dependency
        extractFMpat fs (DSeriesPat s ts) = 
            (map (\t -> DConstraint [getFeatureByName fs s] [t]) --AND connective at term level
                (selectThisLevelTerms fs ts))
            ++ (map (\t -> DConstraint [getFeatureByName fs t] 
                           [Must (getFeatureByName fs s)]) (allThisLevelTerms ts)) 
        --allThisLevelTerms: list of all feature strings of given term series        
        allThisLevelTerms :: DTermSeries -> [String]
        allThisLevelTerms (DTrmSsingle (DTrm s)) = [s]
        allThisLevelTerms (DTrmSsingle (DTrmPlus s)) = [s]
        allThisLevelTerms (DTrmSsingle (DTrmStar s)) = [s]
        allThisLevelTerms (DTrmSsingle (DTrmBrackets s)) = [s]
        allThisLevelTerms (DTrmS (DTrm s) ss) = s : (allThisLevelTerms ss)
        allThisLevelTerms (DTrmS (DTrmPlus s) ss) = s : (allThisLevelTerms ss)
        allThisLevelTerms (DTrmS (DTrmStar s) ss) = s : (allThisLevelTerms ss)
        allThisLevelTerms (DTrmS (DTrmBrackets s) ss) = s : (allThisLevelTerms ss)
        --selectThisLevelTerms: 
        selectThisLevelTerms :: Set Feature -> DTermSeries -> [FeatureConstraint]
        selectThisLevelTerms fs (DTrmSsingle (DTrm t)) = [Must (getFeatureByName fs t)]
        selectThisLevelTerms fs (DTrmSsingle (DTrmPlus t)) = [Must (getFeatureByName fs t)]
        selectThisLevelTerms fs (DTrmSsingle (DTrmStar t)) = [] --no implications to t
        selectThisLevelTerms fs (DTrmSsingle (DTrmBrackets t)) = [] --no implications to t
        selectThisLevelTerms fs (DTrmS (DTrm t) ts) = 
            (Must (getFeatureByName fs t)) : (selectThisLevelTerms fs ts)
        selectThisLevelTerms fs (DTrmS (DTrmPlus t) ts) = 
            (Must (getFeatureByName fs t)) : (selectThisLevelTerms fs ts)
        selectThisLevelTerms fs (DTrmS (DTrmStar t) ts) = 
            selectThisLevelTerms fs ts --no implications to t
        selectThisLevelTerms fs (DTrmS (DTrmBrackets t) ts) = 
            selectThisLevelTerms fs ts --no implications to t

-- extracts feature model out of a GuiDSL constraints
extractFMfromGuiDSLconstr :: Set Feature -> [DConstr] -> [DomainConstraint]
extractFMfromGuiDSLconstr fs [] = []
extractFMfromGuiDSLconstr fs ((DConstrImpl pre con):cs) = 
    (extractConstrImpl fs pre con) ++ (extractFMfromGuiDSLconstr fs cs)
extractFMfromGuiDSLconstr fs ((DConstrIff pre con):cs) = 
    (extractConstrImpl fs pre con) ++ (extractConstrImpl fs con pre) 
        ++ (extractFMfromGuiDSLconstr fs cs)
 
-- transforms a guiDSL constraint to an internal constraint
-- hopefully, this is possible in a number of steps < n^2
extractConstrImpl :: Set Feature -> DConstrExp -> DConstrExp -> [DomainConstraint]
extractConstrImpl fs pre con = 
    [DConstraint ps cs | ps <- extractPremises fs pre, cs <- extractConclusions fs con]
    where
    extractPremises :: Set Feature -> DConstrExp -> [[Feature]]
    extractPremises fs (F s) = [[(getFeatureByName fs s)]]
    extractPremises fs (ExpNot _) = --TODO if needed, the internal model is to be enhanced
        error "Cannot translate 'not' in premises of feature constraints"
    extractPremises fs (ExpAnd e1 e2) = --exponential number of combinations of lists
        concatMap (\p -> map (p ++) (extractPremises fs e2)) (extractPremises fs e1)
    extractPremises fs (ExpOr e1 e2) = --makes several lists
        (extractPremises fs e1) ++ (extractPremises fs e2)
    extractConclusions :: Set Feature -> DConstrExp -> [[FeatureConstraint]]
    extractConclusions fs (F s) = [[Must (getFeatureByName fs s)]]
    extractConclusions fs (ExpNot e) = extractConclusionsInv fs e
    extractConclusions fs (ExpAnd e1 e2) = --makes several lists
        (extractConclusions fs e1) ++ (extractConclusions fs e2)
    extractConclusions fs (ExpOr e1 e2) = --exponential number of combinations of lists
        concatMap (\p -> map (p ++) (extractConclusions fs e2)) (extractConclusions fs e1)
    extractConclusionsInv :: Set Feature -> DConstrExp -> [[FeatureConstraint]] --with negation
    extractConclusionsInv fs (F s) = [[MustNot (getFeatureByName fs s)]]
    extractConclusionsInv fs (ExpNot e) = extractConclusions fs e
    extractConclusionsInv fs (ExpAnd e1 e2) = --exponential number of combinations of lists
        concatMap (\ps -> map ((++) ps) (extractConclusionsInv fs e2)) (extractConclusionsInv fs e1)
    extractConclusionsInv fs (ExpOr e1 e2) = --makes several lists
        (extractConclusionsInv fs e1) ++ (extractConclusionsInv fs e2)

-- TODO: Perhaps we need Tree Feature instead of DomainConstraint.
--       This will be a serious amount of work, but possible.
--       Most occurrences of DomainConstraint will have to be replaced.


-- extracts features from domain constraints. Note that this function demands 
-- that for every unique feature f there is a constraint (f->_)
-- running time: O (c + f), where c=#constraints, f=#features
extractFeaturesFromGuiDSL :: [DProduction] -> Set Feature
extractFeaturesFromGuiDSL ps = Data.Set.map 
    (\s -> Feature { featName=s, 
                     elements=Data.Set.empty, 
                     references=Data.Set.empty, 
                     contracts=Data.Set.empty }) 
    (extractNamesFromGuiDSL ps)

extractNamesFromGuiDSL :: [DProduction] -> Set String
extractNamesFromGuiDSL []       = Data.Set.empty
extractNamesFromGuiDSL ((DProd f patterns):ps) = 
    Data.Set.insert f (Data.Set.union (extractNamesHlpPs patterns)
        (extractNamesFromGuiDSL ps))
    where 
        extractNamesHlpPs :: DPatterns -> Set String
        extractNamesHlpPs (DPatn p) = extractNamesHlpP p
        extractNamesHlpPs (DPatns p ps) = 
            Data.Set.union (extractNamesHlpP p) (extractNamesHlpPs ps)
        extractNamesHlpP :: DPattern -> Set String
        extractNamesHlpP (DSeriesPat name ts) = 
            Data.Set.insert name (extractNamesHlpTs ts)
        extractNamesHlpP (DSinglePat (DOptTerm f)) = Data.Set.singleton f
        extractNamesHlpTs :: DTermSeries -> Set String
        extractNamesHlpTs (DTrmS t ts) = 
            Data.Set.insert (extractNamesHlpT t) (extractNamesHlpTs ts)
        extractNamesHlpTs (DTrmSsingle t) = Data.Set.singleton (extractNamesHlpT t)
        extractNamesHlpT :: DTerm -> String
        extractNamesHlpT (DTrm s) = s
        extractNamesHlpT (DTrmPlus s) = s
        extractNamesHlpT (DTrmStar s) = s
        extractNamesHlpT (DTrmBrackets s) = s


-- FUJI: INTRODUCES + REF + COMPOSEORDER

-- input: .introduces file with feature to element correspondence
-- input: file containing features in composition order (base feature first)
-- input: .rsf reference information file name
-- output: set of features with references; list of feature names in composition order
constructFujiFeatures :: String -> String -> String -> (FeatureSet, [String])
constructFujiFeatures fFile cFile rFile = (Data.Set.fromList (Data.Map.elems 
    (insertRefs (insertElems (generateFeatures composeOrderResult) introResult) rsfResult)),
    composeOrderResult)
    where
        -- .introduces file: list of elements with the feature name they belong to
        introResult :: [(String, Element)]
        introResult = (Introduces.parseElements . Introduces.lexer 
            . Introduces.removeComments . unsafePerformIO) (readFile fFile)
        -- composition file: list of feature names in order of composition (base feature first)
        composeOrderResult :: [String]
        composeOrderResult = (CompOrder.extractCompositionOrder . CompOrder.lexer 
            . CompOrder.removeComments . unsafePerformIO) (readFile cFile)
        -- .RSF file: list of references with feature name the reference source belongs to
        rsfResult :: [ReferenceExpl]
        rsfResult = (RSF.removeReferencesToJavaStdLib . RSF.parseRSF . RSF.lexer 
            . RSF.removeComments . unsafePerformIO) (readFile rFile)
        -- interprets a given list of strings as a set of empty features
        generateFeatures :: [String] -> Map String Feature
        generateFeatures ss = Data.Map.fromList (map (\s -> 
            (s, Feature {featName=s,elements=Data.Set.empty,
                         references=Data.Set.empty,contracts=Data.Set.empty})) ss)

-- inserts given elements into a map of features
-- #1: map of (maybe) empty features (contents are conserved)
-- #2: list of elements with feature name they belong to
-- result: #1 set of features with #2 elements
-- error: if an #2 element belongs to a feature name not in #1
insertElems :: Map String Feature -> [(String, Element)] -> Map String Feature
insertElems fs [] = fs
insertElems fs ((n,e):es) = case Data.Map.lookup n fs of
                Just f -> insertElems
                    (Data.Map.insert 
                          n
                          (Feature {featName=n, 
                                    elements=Data.Set.insert e (elements f),
                                    references=references f,
                                    contracts=contracts f})
                          fs) es
                _ -> error ("insertElems: Did not find feature name '" ++ n 
                            ++ "' in features: " ++ show (Data.Map.keys fs))

-- inserts given references into a map of features
-- #1: map of (maybe) empty features (contents are conserved)
-- #2: list of references with feature name the sources belong to
-- result: #1 set of features with #2 references
-- error: if an #2 source element belongs to a feature name not in #1
insertRefs :: Map String Feature -> [ReferenceExpl] -> Map String Feature
insertRefs fs [] = fs
insertRefs fs (((n,e1),e2):rs) = case Data.Map.lookup n fs of
                Just f -> insertRefs
                    (Data.Map.insert 
                          n
                          (Feature {featName=n, 
                                    elements=elements f,
                                    references=Data.Set.insert (e1,e2) (references f),
                                    contracts=contracts f})
                          fs) rs
                _ -> error ("insertRefs: did not find feature name '" ++ n 
                            ++ "' in features: " ++ show (Data.Map.keys fs))

-- sorts features in the order of composition 
-- input: feature names in compose sequence (first one is first)
-- input: map of features with elements
-- output: list of features in sequence of composition (first one is first)
-- running time: n(log n) [n=#features]
sortFeaturesComposeOrder :: ([String], Map String Feature) -> [Feature]
sortFeaturesComposeOrder ([],_) = []
sortFeaturesComposeOrder ((l:ls),fs) = let lLookup = Data.Map.lookup l fs in
    if (lLookup == Nothing)
    then error ("While sorting features by the order of composition: Feature '"
               ++ l ++ "' is missing in Map.")
    else (fs Data.Map.! l) : (sortFeaturesComposeOrder (ls,fs))

-- sorts a given feature set according to a list of names
-- #1: sorted list of names (all #2's feature names must be in this list!)
-- #2: set of features to sort
-- note: if a feature in #2 is not represented in #1, it will be missing in the result!
-- running time: n(log n) [n=#features]
sortFeatures2ComposeOrder :: [String] -> FeatureSet -> [Feature] 
sortFeatures2ComposeOrder []     _  = []
sortFeatures2ComposeOrder (n:ns) fs = if (Data.Set.null intersec) 
    then sortFeatures2ComposeOrder ns fs
    else (Data.Set.findMin intersec) : (sortFeatures2ComposeOrder ns fs)
    where intersec = Data.Set.intersection fs (Data.Set.singleton 
              Feature{featName=n,
                      elements=Data.Set.empty,
                      references=Data.Set.empty,
                      contracts=Data.Set.empty})

-- Occasionally, FeatureHouse does not bring correct feature correspondences in
-- reference declarations (.rsf files).
-- This function corrects the problem by assigning that feature to a
-- reference which has the element mentioned and is next in composition order
-- input: list of features in sequence of composition (*last* one is first)
-- output: same as input, but with corrected feature correspondence
--         also sets the last implementing feature if there is an "" in #2
correctReferences :: [Feature] -> [ReferenceExpl] -> ([Feature],[ReferenceExpl])
correctReferences fs []     = (fs,[])
correctReferences fs (r:rs) = (fs, (corrFeatureRef r fs) : (snd(correctReferences fs rs)))
    where corrFeatureRef :: ReferenceExpl -> [Feature] -> ReferenceExpl
          corrFeatureRef r fs = corrFeatureRefL r fs
          --remember that only fs have complete elements, f1/f2 are only Strings
          --find f by cropping heads from fs, checking names
          --look if found f has e, if not, look in tail fs
          --return reference with correct feature name
          corrFeatureRefL :: ReferenceExpl -> [Feature] -> ReferenceExpl
          corrFeatureRefL ((f,e),r) fs = corrFeatureRefL1 ((f,e),r) (cropUntil fs f)
          corrFeatureRefL1 :: ReferenceExpl -> [Feature] -> ReferenceExpl
          corrFeatureRefL1 ((f,e),r) [] = error ("no feature found implementing reference source '"
              ++(show e)++"' with type '"++(elType e)++"' - should be in feature '"++f++"' or earlier in composition (exact type needed)")
          corrFeatureRefL1 ((f,e),r) (f1:fs) = if (Data.Set.member e (elements f1))
              then ((featName f1,e),r) else (corrFeatureRefL1 ((f,e),r) fs)

-- crops heads from feature list until name matches
-- input: feature list
-- input: name of feature to crop heads until
-- output: sublist of #1, with cropped heads, maybe #1 itself
cropUntil :: [Feature] -> String -> [Feature]
cropUntil [] f = error ("could not correct feature correspondence to references: "
                           ++"did not find feature '"++(show f)++"' in reference")
cropUntil (f1:fs) f = if (f==featName f1 || f=="") then (f1:fs) else cropUntil fs f


-- extracts features from domain constraints. Note that this function demands 
-- that for every unique feature f there is a constraint (f->_)
-- running time: O (c + f), where c=#constraints, f=#features
extractFeatures :: [DomainConstraint] -> Set Feature
extractFeatures []       = Data.Set.empty
extractFeatures (dc:dcs) = 
    Data.Set.union (extractFeaturesHlp dc) (extractFeatures dcs)
    where extractFeaturesHlp :: DomainConstraint -> Set Feature
          extractFeaturesHlp (DConstraint [] _) = Data.Set.empty
          extractFeaturesHlp (DConstraint (f:fs) _) = Data.Set.singleton f




-- reads .rsf and guidsl-.model files and
-- compares the extracted domain and referential models and prints results
-- #1: features.rsf - which feature introduces which elements
--     note: must contain all features mentioned in #2 and #3
-- #2: .rsf - file (produced by Doxygen) contains implementation references
-- #3: guidsl .model - the domain model is constructed from this file
--     note: must contain all features mentioned in #1 (?? not tested)
-- #4: file with all feature names in composition order (base feature first)
printProductLineAbstractness :: String -> String -> String -> String -> IO ()
printProductLineAbstractness ff rf mf cf = do
    putStrLn "- full abstractness of model: "
    catch (do
           putStrLn (if (fst abstractness) 
                     then "feature model is fully abstract."
                     else "erroneus feature selection:")
--           when (not (fst abstractness)) (putStrLn$("some unsupplied references: "++)$show$ 
--               (getUnsuppliedReferences$fromJust$snd$abstractness)))
           when (not (fst abstractness)) 
               (do
               putStr$"    "
               putStrLn$ (show $ sortFeatures2ComposeOrder (snd rFeatures) 
                         (Data.Set.fromList $ fromJust $ snd abstractness))
               mapM_ (\r -> putStrLn$( --a line for every reference, ignoring results
                   "dangling reference:"++('\n':"    ")++ show (snd r) 
                   ++('\n':"    in feature '") ++ show (fst r) ++ "'"
                   ++('\n':"features that provide proper targets:")++('\n':"    ") 
                   ++ show(Data.Set.toList(getFeaturesWithElement (fst rFeatures) (snd$snd r))) 
                   ))
                   (getUnsuppliedReferences (fromJust$snd$abstractness) (fst rFeatures))
               )
        ) (\e-> putStrLn $ show e)
--    putStr "- model maximality: "
--    catch (do 
--           putStrLn$show$ fst maximality 
--           when (not (fst maximality)) (do
--               putStrLn$ "    counterexample: "++(show $ fromJust $ snd maximality)
--               )
--           )
--        (\e-> putStrLn $ show e)
    where
        -- features with references
        rFeatures :: (FeatureSet, [String])
        rFeatures = constructFujiFeatures ff cf rf
        -- domain product line (does not support guidsl "hidden features")
        dPl :: ProductLine
        dPl = guiDSLtoProductLine (readGuiDSLmodel mf)
        abstractness = checkFeatureModelFullyAbstract (fst rFeatures, snd dPl)


-- reads .rsf and guidsl-.model files and
-- compares the extracted domain and single references and prints results
-- #1: features.rsf - which feature introduces which elements
--     note: must contain all features mentioned in #2 and #3
-- #2: .rsf - file (produced by Doxygen) contains implementation references
-- #3: guidsl .model - the domain model is constructed from this file
--     note: must contain all features mentioned in #1 (?? not tested)
-- #4: file with all feature names in composition order (base feature first)
printReferentialAbstractness :: String -> String -> String -> String -> IO ()
printReferentialAbstractness ff rf mf cf = do 
    putStrLn ("full abstractness of model to each reference ("++(show $ length refs)++"):")
    catch (mapM_ (\ref -> printReferenceAbstractness 
                          (fst rFeatures, snd dPl) (snd rFeatures) ref) refs) 
        (\e-> putStrLn $ show e)
    putStrLn ""
    where
        -- features with references
        rFeatures :: (FeatureSet, [String])
        rFeatures = constructFujiFeatures ff cf rf
        -- domain product line (does not support guidsl "hidden features")
        dPl :: ProductLine
        dPl = guiDSLtoProductLine (readGuiDSLmodel mf)
        -- extracts all references from features and makes the source feature explicit
        refs :: [ReferenceExpl]
        refs = concatMap (\f -> map (\r -> ((featName f,fst r), snd r)) 
                                    (Data.Set.toList (references f))) (Data.Set.toList (fst rFeatures))

-- prints abstractness of model in respect to one reference
-- #1: product line with features with elements, and feature model to be checked
-- #2: single reference to check
printReferenceAbstractness :: ProductLine -> [String] -> ReferenceExpl -> IO ()
printReferenceAbstractness pl co ref = (do 
        putStr "."
        when (not (fst abstractnessRef)) (do
            putStrLn "*"
            putStrLn$ "erroneus feature selection:\n    "++(show $ sortFeatures2ComposeOrder co 
                           (Data.Set.fromList $ fromJust $ snd abstractnessRef))
            putStr ("dangling reference:\n    ("
                ++(show(snd(fst(ref))))++", "++(show(snd(ref)))
                ++")\n    in feature '"++(show(fst(fst(ref))))++"'")
            putStrLn$( --a line for every reference, ignoring results
                "\nfeatures that provide proper targets:\n    " 
                ++ show(Data.Set.toList(getFeaturesWithElement (fst pl) (snd ref))))
            )
        ) 
    where
        abstractnessRef :: (Bool, Maybe [Feature])
        abstractnessRef = checkReferenceFullyAbstract (fst pl, snd pl) ref

-- try:
-- showGuiDSLmodel "model.m"
-- readGuiDSLmodel "model.m"
-- guiDSLtoProductLine (readGuiDSLmodel "model.m")
-- buildAllConfigurations (guiDSLtoProductLine (readGuiDSLmodel "model.m"))
-- checkFeatureModelSatisfiable (guiDSLtoProductLine (readGuiDSLmodel "model.m"))

-- constructRSFfeatures "GPLComp.features.rsf" "GPLComp.rsf"
-- readRSFextractFeatures "GPLComp.features.rsf"
-- readRSFreferences "GPLComp.rsf"

-- checkFeatureModelFullyAbstract (constructRSFfeatures "GPL/GPLComp.features.rsf" "GPL/GPLComp.rsf", snd(guiDSLtoProductLine (readGuiDSLmodel "GPL/GPL.model")))
-- checkFeatureModelMaximal (constructRSFfeatures "GPL/GPLComp.features.rsf" "GPL/GPLComp.rsf", snd(guiDSLtoProductLine (readGuiDSLmodel "GPL/GPL.model")))
-- getUnsuppliedReferences $ fromJust $ snd $ checkFeatureModelFullyAbstract (constructRSFfeatures "GPL/GPLComp.features.rsf" "GPL/GPLComp.rsf", snd(guiDSLtoProductLine (readGuiDSLmodel "GPL/GPL.model")))

-- printProductLineAbstractness "GPLfull2/GPLComp.features.rsf" "GPLfull2/GPLComp.rsf" "GPLfull2/GPL.model"
-- printProductLineAbstractness "GPLmini/GPLComp.features.rsf" "GPLmini/GPLComp.rsf" "GPLmini/GPL.model"
-- printProductLineAbstractness "email/features.exp.rsf" "email/EmailSystemComp.rsf" "email/model.m" 

-- GPL
-- printProductLineAbstractness "GPL/GPLcompAll.introduces" "GPL/GPLcompAll.rsf" "GPL/GPL.model" "GPL/GPL.comporder"
-- printReferentialAbstractness "GPL/GPLcompAll.introduces" "GPL/GPLcompAll.rsf" "GPL/GPL.model" "GPL/GPL.comporder"


-- TANKWAR
-- printProductLineAbstractness "TankWar/TankWarComp.introduces" "TankWar/TankWarComp.rsf" "TankWar/TankWar.model" "TankWar/TankWarCompAllFuji.features"
-- printReferentialAbstractness "TankWar/TankWarCompAll.introduces" "TankWar/TankWarCompAll.rsf" "TankWar/TankWar.model" "TankWar/TankWarCompAllFuji.features"

-- VIOLET
-- printProductLineAbstractness "Violet/VioletCompAll.introduces" "Violet/VioletCompAll.rsf" "Violet/Violet.model" "Violet/VioletCompAll.features"
-- printReferentialAbstractness "Violet/VioletCompAll.introduces" "Violet/VioletCompAll.rsf" "Violet/Violet.model" "Violet/VioletCompAll.features"

