{
------------------------------------------------
-- FeatureTweezer module: Feature-Element-Parser
-- 2010-09 Wolfgang Scholz
------------------------------------------------
-- Parser for extracting element names with corresponding features 
-- out of *.introduces files
--
module Introduces where
import Char
import FeatureStructure


}

%name parseElements
-- [Token] -> [(String, Element)]
-- Result will be a list of tuples:
-- a list of Elements and the Features they reside in
%tokentype { Token }
%error { parseError }

%token
    Introduces      { TokenIntroduces }
    sep             { TokenSeparator } -- ::
    brkLeft         { TokenBrkLeft }   -- (
    brkRight        { TokenBrkRight }  -- )
    comma           { TokenComma }     -- ,
    var             { TokenElem $$ }

--%left '('
--%left not
%%

IntroLines :: { [(String, Element)] }
    : IntroLine IntroLines  { ($1:$2) }
    | IntroLine             { ([$1]) }

IntroLine :: { (String, Element) }
    : Introduces var IntroElem { ($2, $3) }

IntroElem :: { Element }
    : sep var
          { Element{elName=$2,elType="Field",value=""} }
    | var sep var 
          { Element{elName=$1++('.':$3),elType="Field",value=""} }
    | sep var brkLeft brkRight
          { Element{elName=$2,elType="Method:void",value=""} }
    | var sep var brkLeft brkRight
          { Element{elName=$1++('.':$3),elType="Method:void",value=""} }
    | sep var brkLeft RSFParams brkRight
          { Element{elName=$2,elType="Method:"++$4,value=""} }
    | var sep var brkLeft RSFParams brkRight
          { Element{elName=$1++('.':$3),elType="Method:"++$5,value=""} }

RSFParams :: { String }
    : var comma RSFParams { $1 ++ (',':$3) }
    | var                 { $1 }


{

data Token
      = TokenIntroduces
      | TokenSeparator    -- ::
      | TokenBrkLeft      -- (
      | TokenBrkRight     -- )
      | TokenComma        -- ,
      | TokenElem String

instance Show Token where
    show TokenIntroduces = "introduces"
    show TokenSeparator  = "::"
    show TokenBrkLeft = "("
    show TokenBrkRight  = ")"
    show TokenComma = ","
    show (TokenElem str) = str

parseError :: [Token] -> a
parseError [] = error "Parse error"
parseError (x:xs) = error ("'Introduces' parse error before : \"" 
    ++ (foldr (++) "" (map (' ':) (map show (take 12 (x:xs))))) 
    ++ "\"")
-- lexer

lexer :: String -> [Token]
lexer [] = []
lexer (c:cs) 
      | isSpace c = lexer cs
      | c=='"' = lexVarQuote cs
      | c=='<' = lexSkipUntilRightWedge cs
      | c=='(' = TokenBrkLeft : lexer cs
      | c==')' = TokenBrkRight : lexer cs
      | c==',' = TokenComma : lexer cs
      | c==':' = if (head cs ==':') then TokenSeparator : lexer (tail cs)
                 else error ("single ':' occurred before "  ++ take 24 (cs))
      | otherwise = lexVar (c:cs)

-- Stream is tokenized at:
-- <space> , ( ) :
-- Note that there may be no token with ':' in it (but it can be quoted with '"')
lexVar cs =
   case span 
          (\x -> (not(isSpace x) && x/=',' && x/='(' && x/=')' && x/=':' && x/='<')) 
          cs of
      ("introduces", rest) -> (TokenIntroduces) : lexer rest
      ("", rest)   -> error ("Lexer error: cannot eliminate \"" ++ take 1 (rest) ++ "\"")
      (var,rest)   -> (TokenElem var) : lexer rest

-- note that '"' may not be part of any reserved word
lexVarQuote cs = 
    case span (\x -> x/='"') cs of 
        (var,rest)  ->  (TokenElem var) : lexer (tail rest)

-- Generics ( <?> ) are just skipped
lexSkipUntilRightWedge cs = 
    case span (\x -> x/='>') cs of 
        (var,rest)  ->  lexer (tail rest)

-- discards comments
removeComments :: String -> String
removeComments [] = ""
removeComments ('#':xs) = discardLine xs
    where discardLine :: String -> String --until EOL
          discardLine [] = ""
          discardLine (x:xs) | isPrint x = discardLine xs
                             | otherwise = removeComments xs
removeComments (x:xs) = x : (removeComments xs)

main = getContents >>= print . parseElements . lexer . removeComments 

}

--happy example.y 
--happy example.y -i 



