{
------------------------------------------
-- FeatureTweezer module: RSF-Parser
-- 2010-06 Wolfgang Scholz
------------------------------------------
-- Parser for extracting references out of *.rsf files

module RSF where
import Char
import FeatureStructure
import Data.List

-- Fuji does not discern between references inside the project and references, 
-- which have their target in external libraries. This function filters
-- references to the java standard library, which occur most often in java
-- projects.
-- #1: parsed references
-- output: #1 without references to the java standard library (java.*) 
-- running time: linear
removeReferencesToJavaStdLib :: [ReferenceExpl] -> [ReferenceExpl]
removeReferencesToJavaStdLib [] = []
removeReferencesToJavaStdLib ((s,t):rs) 
    | Data.List.isPrefixOf "java." (elName t)  = removeReferencesToJavaStdLib rs
    | Data.List.isPrefixOf "javax." (elName t) = removeReferencesToJavaStdLib rs
    | Data.List.isPrefixOf "@primitive" (elName t) = removeReferencesToJavaStdLib rs
    | otherwise                                = (s,t) : (removeReferencesToJavaStdLib rs)

}

%name parseRSF
-- [Token] -> [ReferenceExpl]
-- Result will be a list of references, the source element with feature it occurs in
%tokentype { Token }
%error { parseError }

%token
    REF             { TokenREF }       -- REF
    ORI             { TokenORI }       -- ORI
    sep             { TokenSeparator } -- ::
    brkLeft         { TokenBrkLeft }   -- (
    brkRight        { TokenBrkRight }  -- )
    comma           { TokenComma }     -- ,
    var             { TokenElem $$ }   -- any String (. is no separator)

--%left '('
--%left not
%%

RSFLines :: { ([ReferenceExpl]) }
    : RSFLine RSFLines  { $1 : $2 }
    | RSFLine           { [$1] }

RSFLine :: { ReferenceExpl }
    : REF RSFElemExpl RSFElem { ((fst($2), snd($2)), $3) }
    | ORI RSFElemExpl RSFElem { ((fst($2), snd($2)), $3) }
-- ^^ Note that there is no target feature field in ReferenceExpl
--    A search over all features must be conducted to find the target Element
--    It has to be assured that any found elements satisfies the dependency
-- TODO: There might be a need for special treatment of original()-references

-- need separators: <space> :: ( ) ,


RSFElemExpl :: { (String, Element) }
    :  var sep RSFElem { ($1, $3) } 

RSFElem :: { Element }
    : sep var
          { Element{elName=$2,elType="Field",value=""} }
    | var sep var 
          { Element{elName=$1++('.':$3),elType="Field",value=""} }
    | sep var brkLeft brkRight
          { Element{elName=$2,elType="Method:void",value=""} }
    | var sep var brkLeft brkRight
          { Element{elName=$1++('.':$3),elType="Method:void",value=""} }
    | sep var brkLeft RSFParams brkRight
          { Element{elName=$2,elType="Method:"++$4,value=""} }
    | var sep var brkLeft RSFParams brkRight
          { Element{elName=$1++('.':$3),elType="Method:"++$5,value=""} }

RSFParams :: { String }
    : var comma RSFParams { $1 ++ (',':$3) }
    | var                 { $1 }

{

data Token
      = TokenREF          -- REF
      | TokenORI          -- ORI
      | TokenSeparator    -- ::
      | TokenBrkLeft      -- (
      | TokenBrkRight     -- )
      | TokenComma        -- ,
      | TokenElem String

instance Show Token where
    show TokenREF  = "REF"
    show TokenORI  = "ORI"
    show TokenSeparator  = "::"
    show TokenBrkLeft = "("
    show TokenBrkRight  = ")"
    show TokenComma = ","
    show (TokenElem str) = str

parseError :: [Token] -> a
parseError [] = error "Parse error"
parseError (x:xs) = error ("RSF Parse error before : \"" 
    ++ (foldr (++) "" (map (' ':) (map show (take 12 (x:xs))))) 
    ++ "\"")
    
--unBrk :: String -> String
--unBrk [] = error "unBrk: lost brackets"
--unBrk "()" = ""
--unBrk (x:xs) = x:(unBrk xs)
--
--unBrk2 :: String -> String -- only to discern from unBrk
--unBrk2 [] = "" --no error here
--unBrk2 "()" = ""
--unBrk2 (x:xs) = x:(unBrk2 xs)

-- lexer

lexer :: String -> [Token]
lexer [] = []
lexer (c:cs) 
      | isSpace c = lexer cs
      | c=='"' = lexVarQuote cs
      | c=='<' = lexSkipUntilRightWedge cs
      | c=='(' = TokenBrkLeft : lexer cs
      | c==')' = TokenBrkRight : lexer cs
      | c==',' = TokenComma : lexer cs
      | c==':' = if (head cs ==':') then TokenSeparator : lexer (tail cs)
                 else error ("single ':' occurred before "  ++ take 24 (cs))
      | otherwise = lexVar (c:cs)

-- Stream is tokenized at:
-- <space> , ( ) :
-- Note that there may be no token with ':' in it (but it can be quoted with '"')
lexVar cs =
   case span 
          (\x -> (not(isSpace x) && x/=',' && x/='(' && x/=')' && x/=':' && x/='<')) 
          cs of
      ("REF",rest) -> TokenREF : lexer rest
      ("ORI",rest) -> TokenORI : lexer rest
      ("", rest)   -> error ("Lexer error: cannot eliminate \"" ++ take 1 (rest) ++ "\"")
      (var,rest)   -> (TokenElem var) : lexer rest

-- note that '"' may not be part of any reserved word
lexVarQuote cs = 
    case span (\x -> x/='"') cs of 
        (var,rest)  ->  (TokenElem var) : lexer (tail rest)

-- Generics ( <?> ) are just skipped
lexSkipUntilRightWedge cs = 
    case span (\x -> x/='>') cs of 
        (var,rest)  ->  lexer (tail rest)

-- discards comments
removeComments :: String -> String
removeComments [] = ""
removeComments ('#':xs) = discardLine xs
    where discardLine :: String -> String --until EOL
          discardLine [] = ""
          discardLine (x:xs) | isPrint x = discardLine xs
                             | otherwise = removeComments xs
removeComments (x:xs) = x : (removeComments xs)

main = getContents >>= print . removeReferencesToJavaStdLib . parseRSF . lexer . removeComments 

}

--happy example.y 
--happy example.y -i 



